"""Loops for Bayesian quadrature."""

# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from .bayesian_monte_carlo_loop import BayesianMonteCarlo  # noqa: F401
from .vanilla_bq_loop import VanillaBayesianQuadratureLoop  # noqa: F401
from .wsabil_loop import WSABILLoop  # noqa: F401

__all__ = [
    "BayesianMonteCarlo",
    "VanillaBayesianQuadratureLoop",
    "WSABILLoop",
    "point_calculators",
]
